###
### "THE BEER-WARE LICENSE":
### Alberto Ramos wrote this file. As long as you retain this 
### notice you can do whatever you want with this stuff. If we meet some 
### day, and you think this stuff is worth it, you can buy me a beer in 
### return. <alberto.ramos@cern.ch>
###
### file:    ScalarForce.jl
### created: Wed Oct  6 15:39:07 2021
###                               

function krnl_force_scalar!(fgauge, fscalar, U::AbstractArray{TG}, Phi::AbstractArray{TS}, sp::ScalarParam{NP,T}, lp::SpaceParm{N,M,D}) where {TG,TS,NP,T,N,M,D}

    b, r = CUDA.threadIdx().x, CUDA.blockIdx().x

    Ush = @cuStaticSharedMem(TG, (D,N))
    Psh = @cuStaticSharedMem(TS, (D,NP))
    
    for id in 1:N
        Ush[b,id] = U[b,id,r]
    end
    for i in 1:NP
        Psh[b,i] = Phi[b,i,r]
    end
    sync_threads()

    for n in 1:NP
        fscalar[b,n,u] = zero(TS)
        for id in 1:N
            bu, ru = up((b,r), id, lp)
            bd, rd = dw((b,r), id, lp)

            if ru == r
                p1 = Ush[b,id]*Psh[bu,n]
            else
                p1 = Ush[b,id]*Phi[bu,n,ru]
            end

            if rd == r
                fscalar[b,n,u] += (-2*sp.kap[n])*(p1 + dag(Psh[bd,n])*Ush[bd,id])
            else
                fscalar[b,n,u] += (-2*sp.kap[n])*(p1 + dag(Phi[bd,n,rd])*U[bd,id,rd])
            end
            
            fgauge[b,id,r] += (2*sp.kap[n])*projalg(p1*dag(Psh[b,n]))
        end
        fscalar[b,n,r] += ( 2 + 2*sp.eta[n]*(dot(Psh[b,n],Psh[b,n])-1) ) * Psh[b,n]
    end


    return nothing
end
